/* 
 * Crystal Mud Client
 * Copyright (C) 1998-2002 Abigail Brady
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <resolv.h>

#include "config.h"

using namespace std;

#include "Socket.h"
#include "io.h"
#include "grid.h"
#include "telnet.h"
#include "crystal.h"
#include "scripting.h"

InAddrList::InAddrList() : addrs(0), n_addrs(0) 
{
}

void InAddrList::add(InAddr *a)
{
  addrs = (InAddr**)realloc(addrs, (n_addrs+1)*sizeof(InAddr*));
  addrs[n_addrs] = a;
  n_addrs++;
}

InAddrList::~InAddrList() {
  for (int i=0;i<n_addrs;i++)
    delete addrs[i];
}

class InAddr4 : public InAddr {
  struct sockaddr_in s;
public:
  InAddr4(const struct in_addr *si) {
    s.sin_family = AF_INET;
    s.sin_port = 0;
    s.sin_addr = *si;
  }
  InAddr4(const struct sockaddr_in *si) : s(*si) {
  }
  virtual ~InAddr4() { };
  void set_port(int port) { s.sin_port = htons(port); };
  virtual string tostring() {
    return inet_ntoa(s.sin_addr);
  }
  sockaddr *addr() {
    return (sockaddr*)&s;
  }
  socklen_t addr_len() {
    return sizeof s;
  }
  int af() { return AF_INET; };
};

#ifdef AF_INET6
class InAddr6 : public InAddr {
  struct sockaddr_in6 s;
public:
  InAddr6(const struct in6_addr *si) {
    s.sin6_family = AF_INET6;
    s.sin6_port = 0;
    s.sin6_addr = *si;
  }
  InAddr6(const struct sockaddr_in6 *si) : s(*si) {
  }
  virtual ~InAddr6() { };
  void set_port(int port) { s.sin6_port = htons(port); };
  virtual string tostring() {
    char dest[100];
    inet_ntop(AF_INET6, &s.sin6_addr, dest, 100);
    dest[99] = 0;
    return dest;
  }
  sockaddr *addr() {
    return (sockaddr*)&s;
  }
  socklen_t addr_len() {
    return sizeof s;
  }
  int af() { return AF_INET6; };
};
#endif

InAddr *InAddr::create(const struct sockaddr *addr)
{
  if (addr->sa_family == AF_INET) {
    return new InAddr4((const struct sockaddr_in *)addr);
  }
#ifdef AF_INET6
  if (addr->sa_family == AF_INET6) {
    return new InAddr6((const struct sockaddr_in6 *)addr);
  }
#endif
  return NULL;
}

InAddrList *InAddr::resolv(const char *name) 
{
  struct hostent *he = gethostbyname(name);
#ifdef AF_INET6
  struct hostent *he6 = gethostbyname2(name, AF_INET6);
#endif
  if (!he
#ifdef AF_INET6
 && !he6
#endif
)
    return NULL;

  InAddrList *list = new InAddrList;
#ifdef AF_INET6
  if (he6 && he6->h_addrtype == AF_INET6)
    list->add(new InAddr6((struct in6_addr*)he6->h_addr_list[0]));
#endif
  if (he && he->h_addrtype == AF_INET) {
    list->add(new InAddr4((struct in_addr*)he->h_addr_list[0]));
  }

  return list;
}


Socket::Socket() {
  buffer = "";
  fd = -1;
}

int Socket::connect(InAddr *addr) {
  fd = socket(addr->af(), SOCK_STREAM, 0);
  if (fd == -1)
    return -1;
  
  int value = ::connect(fd, addr->addr(), addr->addr_len());
  
  int er = errno;
  
  if (value == -1) {
    ::close(fd);
    errno = er;
    return -1;
  }
  
  fcntl(fd, F_SETFL, O_NONBLOCK);
  dead = false;
  
  return true;
}

int Socket::read(char *data, int size) {
  if (dead) {
      errno = ECONNRESET;
      return -1;
  }
  int rval = ::read(fd, data, size);
  if (rval == -1) {
      if (errno == EAGAIN || errno == EWOULDBLOCK) {
	  return -1;
      }
      dead = true;
  }
  return rval;
}

int Socket::write(const char *data, int size) {
    if (dead) {
	errno = ECONNRESET;
	return -1;
    }
    int rval = ::write(fd, data, size);
    if (rval == -1) {
	dead = true;
    }
    return rval;
}

Socket::~Socket() {
  if (!dead) {
    shutdown(fd, 2);
    ::close(fd);
  }
}

void Socket::close() {
  shutdown(fd, 2);
  ::close(fd);
  dead = true;
}

int lookup_service(const string &name)
{
  char *e = 0;
  int port = strtol(name.c_str(), &e, 10);
  if (!*e)
    return port;
  
  struct servent *se = getservbyname(name.c_str(), "tcp");
  if (se) {
    return ntohs(se->s_port);
  } else {
    return -1;
  }
}

int count_chars(const char *s, char ch)
{
  int c = 0;
  while (*s) 
    {
      if (*s == ch)
	c++;
      s++;
    }
  return c;
};

url::url(const char *s) 
{
  url &u = *this;
  u.protocol = "telnet";
  u.service = "telnet";

  if (const char *t=strstr(s, "://")) {
    u.protocol = string(s, t-s);
    s = t + strlen("://");
  }

  const char *srv = 0, *lasth = 0;
  
  if (s[0]=='[') {
    // we have a 6 ip address encoded in [numeric] form.
    s++;
    srv = strchr(s, ']');
    if (srv && srv[1]==':') {
      lasth = srv;
      srv = srv + 2;
    } else {
      srv = 0;
    }
  } else if (count_chars(s, ':')>=2) {
    // it is a 6 address with no service at the end - do nothing
  } else {
    srv = strchr(s, ':');
    if (srv) {
      lasth = srv;
      srv++;
    }
  }

  if (lasth) 
    u.hostname = string(s, lasth-s);
  else
    u.hostname = s;

  string newhost = scripting::lookup_host(u.hostname);
  if (newhost.length()) {
    if (newhost.find(':')!=string::npos) {
      int w = newhost.find(':');
      u.service = newhost.substr(w+1);
      u.hostname = newhost.substr(0, w);
    } else {
      u.hostname = newhost;
    }
  }
  if (srv) {
    const char *sl = strchr(srv, '/');
    u.service = sl?string(srv, sl-srv):string(srv);
  }
}
