/*
 * Crystal Mud Client
 * Copyright (C) 2002-2003 Abigail Brady
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

/*
 *
 *   stdin -> charset filter -> esc seq interpreter -> keycode interpreter -> 
 *                            cmd interpreter
 *
 *   read(telnet) -> telnet decoder -> charset filter filter -> tty handler
 */

#ifndef _XOPEN_SOURCE
#define _XOPEN_SOURCE
#endif

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#define TELOPT_MXP 0x5b

#include <term.h>
#include <curses.h>

#include <locale.h>
#include <wchar.h>
#include <langinfo.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <termios.h>
#include <signal.h>
#include <errno.h>
#include <netdb.h>

#include <netinet/in.h>

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>

#include <iconv.h>

#include <list>
#include <vector>
#include <map>
#include <string>
#include <set>

#include <stdarg.h>
#include <ctype.h>

#include <libintl.h>

#undef lines
#undef newline
#undef grid

using namespace std;

#include "Socket.h"

#include "config.h"

#undef SCROLL

//FILE *logfile = 0;

class conn_t;
class grid_t;

void connect(conn_t *, const char *host, int port);
int file_dump(grid_t *, const char * file);

#include "io.h"
#include "grid.h"
#include "telnet.h"
#include "crystal.h"
#include "scripting.h"

mterm tty;

extern "C" int mk_wcwidth(wchar_t ucs);
int real_wcwidth(wchar_t u) {
  int wid = wcwidth(u);
  if (wid>0) {
    return wid;
  }
  wid = mk_wcwidth(u);
  if (wid>0) {
    return wid;
  }
  wid = 1;
  return 1;
}

const struct cell_t blank = {' ', I_NORM, COL_DEFAULT, 0, COL_DEFAULT, 0 };
const struct cell_t blank2 = {0, I_NORM, COL_DEFAULT, 0, COL_DEFAULT, 0 };

conn_t::conn_t(grid_t *gr) {
  cur_grid = grid = gr;
  slave = new grid_t(NULL);
  
  slave->defbc = 0;
  slave->backcol = 0;
  slave->visible = 0;
  
  telnet = 0;

  buffer = L"";
  c = 0;
  
  cutbuffer = L"";
  logfile = 0;

  lp_prompts = 1;
  mud_cset = "ISO-8859-1";

  gr->set_conn(this);
  if (slave)
    slave->set_conn(this);
}

conn_t::~conn_t() {
  if (slave)
    delete slave;
}

int commandmode = 0;
int hardscroll = 0;

cell_t wantbuffer[MAXHEIGHT][MAXWIDTH];
cell_t havebuffer[MAXHEIGHT][MAXWIDTH];

int evillines[MAXHEIGHT] = { 0,  };

int bad_have = 1;

void show_want() {
  
  int realy = -1;
  int realx = -1;

  if (tty.knowscroll) {
    int anyevil = 0;
    for (int i=0;i<tty.HEIGHT;i++) {
      int notsame = 0, onebelow = 1;
      for (int j=0;j<tty.WIDTH;j++) {
	if (wantbuffer[i][j]!=havebuffer[i+1][j]) {
	  onebelow = 0;
	}
	if (wantbuffer[i][j] != havebuffer[i][j]) {
	  notsame = 1;
	}
      }
      if (notsame && onebelow) {
	for (int j=0;j<tty.WIDTH;j++) {
	  havebuffer[i][j] = wantbuffer[i][j];
	  havebuffer[i+1][j] = blank;
	}
	evillines[i] = 0;
	evillines[i+1] = 1;
	printf("\033[40m");
	printf("\033[%i;%ir", i+1, i+2);
	printf("\033[1S");
	printf("\033[r");
	continue;
      }      
    }
  }
  tty.initcol();

  for (int i=0;i<(tty.HEIGHT+1);i++) {
    int wanty = i+1;
    int wantx = 1;
    tty.died = 0;
    int blah = (MAXHEIGHT-tty.HEIGHT)-1;
    int thislen = -1;
    for (int j=0;j<((i==tty.HEIGHT)?tty.WIDTH-1:tty.WIDTH);j++) {
      if (wantbuffer[i][j].ch)
	thislen = j;
    }

    for (int j=0;j<((i==tty.HEIGHT)?tty.WIDTH-1:tty.WIDTH);j++) {
      if (havebuffer[i][j] == wantbuffer[i][j] && !bad_have
      	  && !evillines[i]) {
      	wantx++;
	continue;
      } else {
	if (realy != wanty) {
	  realy = wanty;
	  realx = wantx;
	  printf("\033[%i;%if", wanty, wantx);
	} else if (realx != wantx) {
	  if (wantx==(realx+1)) {
	    printf("\033[C");
	  } else {
	    printf("\033[%i;%if", wanty, wantx);
	  }
	  realx = wantx;
	}
	
	havebuffer[i][j] = wantbuffer[i][j];
	wantx++;
	realx++;
	tty.plonk(wantbuffer[i][j], j>=thislen);
      }
    }
    evillines[i] = 0;
  } 
  bad_have = 0;
}

set<my_wstring> hl_matches;

void conn_t::show_lines_at(int from, int to, int num)
{
  for (int i=0;i<num;i++) {
    if (i+from == grid->row && !(telnet && telnet->charmode)) {
      for (int j=0;j<tty.WIDTH;j++)
	wantbuffer[i+to-1][j] = blank2;
      continue;
    }
    int mw = MIN(tty.WIDTH, grid->get_len(i+from));
    int j;
    if (tty.utf8) mw++;
    for (j=0;j<mw;j++) {
      cell_t g = grid->get(i+from, j);
      wantbuffer[i+to-1][j] = g;
    }

    my_wstring s;
    for (int j=0;j<grid->get_len(i+from);j++) {
      s += grid->get(i+from, j).ch;
    }

    set<my_wstring>::iterator it;
    for (it=hl_matches.begin();it!=hl_matches.end();it++) {
      size_t l = s.find(*it);
      while (l != -1) {
	for (int j=0;j<it->length();j++)
	wantbuffer[i+to-1][j+l].inv = !wantbuffer[i+to-1][j+l].inv;
	l = s.find(*it, l+1);
      }
    }
      
    while (j<tty.WIDTH) {
      wantbuffer[i+to-1][j] = blank2;
      j++;
    }
  }
}

void conn_t::findnext()
{
  for (int i=hardscroll;i<grid->row;i++) {
    my_wstring s;
    for (int j=0;j<grid->get_len(i);j++) {
      s += grid->get(i, j).ch;
    }
    
    set<my_wstring>::iterator it;
    for (it=hl_matches.begin();it!=hl_matches.end();it++) {
      size_t l = s.find(*it);
      while (l != -1) {
	for (int j=0;j<it->length();j++) {
	  hardscroll = i+1;
	  if (hardscroll>(grid->row-tty.HEIGHT))
	    hardscroll = 0;
	  grid->changed = 1;
	  return;
	}
	l = s.find(*it, l+1);
      }
    }
  }
  hardscroll = 0;
  grid->changed = 1;
}

void display_buffer(conn_t *conn) {
  if (info_to_stderr) 
    return;

  grid_t &grid = *conn->grid;

  if (!grid.changed && !conn->slave->changed)
    return;

  const wchar_t *a = conn->buffer.data();
  int alen = conn->buffer.length();

  const int col = conn->c;
  const wchar_t *txt = a;

  int scroll = 0;

  int start = grid.row - tty.HEIGHT;
  if (start < 0) {
    start = 0;
  }

  if (hardscroll)
    start = hardscroll-1;

  if (conn->telnet && conn->telnet->charmode && !commandmode) {
    if (grid.row<tty.HEIGHT)
      conn->show_lines_at(start, tty.HEIGHT-grid.row+1, tty.HEIGHT+1);
    else
      conn->show_lines_at(start, 1, tty.HEIGHT+1);
    show_want();
    printf("\033[%i;%if", tty.HEIGHT+1, grid.col+1);
    return;
  }

  if (grid.changed) {
    if (hardscroll) {
      conn->show_lines_at(start, 1, (tty.HEIGHT-4));
      conn->show_lines_at((grid.row-5)>0?grid.row-5:0, tty.HEIGHT-4, 5);
      for (int i=0;i<tty.WIDTH;i++) {
	wantbuffer[tty.HEIGHT-5][i] = mkchar('=', I_BOLD, COL_WHITE, COL_RED, 0, 0, 0, 0, 0, 0);
      }
    } else {
      if (grid.row<tty.HEIGHT)
	conn->show_lines_at(start, tty.HEIGHT-grid.row+1, tty.HEIGHT);
      else
	conn->show_lines_at(start, 1, tty.HEIGHT);
    }
    grid.changed = 0;
  }

  if (conn->slave && conn->slave->visible) {
    for (int i=0;i<conn->slave->height;i++) {
      for (int j=0;j<conn->slave->width;j++) {
	if (tty.WIDTH-conn->slave->width+j>=0)
	  wantbuffer[i][tty.WIDTH-conn->slave->width+j] = conn->slave->get(i, j);
      }
    }
  }

  cellstring crealprompt;

  if (commandmode) {
    const char *str = _("crystal> ");
    int max = strlen(str);
    while (1) {
      wchar_t c;
      int s=mbtowc(&c, str, max);
      if (s<=0)
	break;
      crealprompt += mkchar(c);
      str += s;
      max -= s;
    }
  } else {
    for (int i=0;i<grid.col;i++) {
      crealprompt += grid.get(grid.row, i);
    }

    if (grid.cstoredprompt.length() && !crealprompt.length()) {
      for (int i=0;i<grid.cstoredprompt.length();i++) {
	crealprompt += grid.cstoredprompt[i];
      }
    }
  }

  int prlen = crealprompt.length();
  
  if (prlen>(tty.WIDTH-20))
    prlen = (tty.WIDTH-20);

  int wid = tty.WIDTH-5;
  if (conn->c+prlen>=wid) {
    scroll = conn->c-wid+1+prlen;
    a += scroll;
    alen -= scroll;
  }
  wid =tty.WIDTH-1;
  
  int i;
  int pl = crealprompt.length();
  if (pl > (tty.WIDTH-20))
    pl = tty.WIDTH-20;

  for (i=0;i<pl;i++)
    wantbuffer[tty.HEIGHT][i] = crealprompt[i];
  
  wid -= prlen;
  if (!(conn->telnet && conn->telnet->allstars) || commandmode) {
    while (alen && wid) {
      if (real_wcwidth(*a)<=wid) {
	wantbuffer[tty.HEIGHT][i++] = mkchar(*a);
	if (real_wcwidth(*a)==2)
	  wantbuffer[tty.HEIGHT][i++] = mkchar(-*a);
      } else
	break;
      wid -= real_wcwidth(*a);
      a++;
      alen--;
    }
  }
  while (wid) {
    wantbuffer[tty.HEIGHT][i++] = blank;
    wid--;
  }
  if (alen) {
    wantbuffer[tty.HEIGHT][i++] = mkchar('>');
  } else
    wantbuffer[tty.HEIGHT][i++] = mkchar(' ');

  show_want();

  int rcol=col;

  for (int i=0;i<col;i++) {
    if (real_wcwidth(txt[i])==2)
      rcol++;
  }

  if ((conn->telnet && conn->telnet->allstars) && !commandmode) 
    rcol = 0;

  printf("\033[%i;%if", tty.HEIGHT+1, rcol-scroll+1+prlen);
  grid.changed = 0;

  fflush(stdout);
}

void wtstring(grid_t *grid, const my_wstring &str) {  
  int len = str.length();
  int i = 0;
  while (len) {
    grid->wterminal(str[i]);
    len--;
    i++;
  }
}

void conn_t::triggerfn(const char *fn) {
  if (strcmp(fn, "12")==0) {
    if (slave) {
      slave->visible = !slave->visible;
      slave->changed = 1;
    }
    return;
  }

  string st = "fn.";
  st += fn;
  st += "\r\n";

  if (telnet)
    telnet->send(st);
}


struct hlist {
  map<int, my_wstring> hist;
  int idx;
  int max;
  hlist() : idx(0), max(0) {
  }

  void insert(const my_wstring &blah) {
    hist[max] = blah;
    max++;
    idx = max;
  }

  bool back() {
    idx--;
    if (idx < 0) {
      idx = 0;
      return 0;
    }
    return 1;
  }

  bool next() {
    idx++;
    if (idx >= max) {
      idx = max;
      return  0;
    }
    return 1;
  }

  const my_wstring &get() {
    static const my_wstring blah = L"";
    if (idx >= max || idx < 0) 
      return blah;
    return hist[idx];
  }
};


hlist cmdhist;

hlist *chist() {
  static hlist hist;
  
  if (commandmode)
    return &cmdhist;

  return &hist;
}

string mks(const my_wstring &w) {
  string s;
  for (int i=0;i<w.length();i++) {
    wchar_t c = w[i];
    if (c < 0x80)
      s += c;
    else
      s += '?';
  }
  return s;
}

void conn_t::dokillword() {
  int orig=c;
  if (c) {
    while(c && isspace(buffer[c-1]))
      c--;
    while(c && !isspace(buffer[c-1]))
      c--;

    my_wstring extra = buffer.substr(orig);
    buffer = buffer.substr(0, c);
    buffer += extra;
  }
}

void conn_t::dodelete() {
  if (c < buffer.length()) {
    buffer.erase(c, 1);
  } else {
    tty.beep();
  }
}

void conn_t::dobackspace()
{
  if (c) {
    buffer.erase(c-1, 1);
    c--;
  } else {
    tty.beep();
  }
}

void conn_t::doscrollstart() {
  if (grid->row < tty.HEIGHT)
    return;
  hardscroll = 1;
  grid->changed = 1;
}

void conn_t::doscrollend() {
  hardscroll = 0;
  grid->changed = 1;
}

void conn_t::doprevhistory() {
  if (chist() && chist()->back()) {
    buffer = chist()->get();
    c = buffer.length();
  }
  else
    tty.beep();
}

void conn_t::donexthistory()
{
  if (chist() && chist()->next()) {
    buffer = chist()->get();
    c = buffer.length();
  }
  else
    tty.beep();
}

void conn_t::doprevchar()
{
  c--;
  if (c < 0) {
    c = 0;
    tty.beep();
  }
}

void conn_t::doprevword()
{
  while (c && !isalnum(buffer[c-1])) {
    c--;
  }
  while (c && isalnum(buffer[c-1])) {
    c--;
  }
}

void conn_t::donextword()
{
  while (c<buffer.length() && !isalnum(buffer[c])) {
    c++;
  }
  while (c<buffer.length() && isalnum(buffer[c])) {
    c++;
  }
}

void conn_t::donextchar()
{
  c++;
  if (c>buffer.length()) {
    c = buffer.length();
    tty.beep();
  }
}

void conn_t::dofirstchar() {
  c = 0;
}

void conn_t::doclearline() {
  buffer = L"";
  c = 0;
}

void conn_t::dolastchar() {
  c = buffer.length();
}

void conn_t::dorefresh() {
  grid->changed=1;
  bad_have = 1;
}

void conn_t::dotranspose() {
  if (c>1 && buffer.length()==c) {
    wchar_t tmp = buffer[c-1];
    buffer[c-1] = buffer[c-2];
    buffer[c-2] = tmp;
    grid->changed = 1;
  } else if (c>0) {
    wchar_t tmp = buffer[c];
    buffer[c] = buffer[c-1];
    buffer[c-1] = tmp;
    c++;
    grid->changed = 1;
  } else {
    tty.beep();
  }
}

void conn_t::doinsertchar(wchar_t ch)
{
  if (c == buffer.length()) {
    buffer += ch;
  } else {
    buffer = buffer.substr(0, c) + ch + buffer.substr(c);
  }
  grid->changed = 1;
  c++;
}

void cmd_quit(conn_t *conn, int &quit) {
  quit = 1;
}

void cmd_z(conn_t *conn, int &quit)
{
  kill(0, SIGTSTP);

  struct termios ti;
  cfmakeraw(&ti);
  tcsetattr(0, TCSADRAIN, &ti);
  
  tty.grabwinsize();

  bad_have = 1;
  conn->grid->changed = 1;
  show_want();
}

void cmd_close(conn_t *conn, int &quit)
{
  if (!conn->telnet) {
    info(conn->grid, _("/// you aren't connected.\n"));
    return;
  }

  delete conn->telnet;
  conn->telnet = 0;
  info(conn->grid, _("/// connection closed.\n"));
}

void cmd_reload(conn_t *conn, int &quit)
{
  scripting::start();
  info(conn->grid, _("/// scripting restarted.\n"));
}

void cmd_help(conn_t *conn, int &quit) {
  info(conn->grid,
       _("/// connect <host> <port>  -- connect to a given host/port\n"
       "/// close                  -- close the current connection\n"
       "/// quit                   -- exit\n"
       "/// z                      -- suspend\n"
       "/// dump <file>            -- dump the backbuffer to a file. appends.\n"
       "/// log <file>             -- sets logfile as specified. appends.\n"
       "/// charset <charset>      -- interprets data from the mud as given charset\n"));
#ifdef MCCP
  info(conn->grid,
       _("/// compress               -- show MCCP compression status \n"));
#endif
}

#ifdef MCCP
void cmd_compress(conn_t *conn, int &quit) {
  if (conn->telnet &&
      conn->telnet->mc && 
      mudcompress_compressing(conn->telnet->mc)) {
    unsigned long compread, uncompread;
    mudcompress_stats(conn->telnet->mc, &compread, &uncompread);
    infof(conn->grid, _("/// wire:%li bytes data:%li bytes.\n"), compread, uncompread);
  } else {
    info(conn->grid, _("/// no compression\n"));
  }
}
#endif

void cmd_connect(conn_t *conn, int &quit)
{
  my_wstring host = conn->buffer;
  conn->buffer = L"";
  conn->c = 0;
  host = host.substr(host.find(' ')+1);

  my_wstring port;

  if (host.find(' ')!=string::npos) {
    port = host.substr(host.find(' ')+1);
    host = host.substr(0, host.find(' '));
  }
  
  string cport = mks(port);
  string chost = mks(host);

  url u = url(chost.c_str());
  if (cport.length()!=0) {
    u.service = cport;
  }
  
  if (u.protocol != "telnet") {
    infof(conn->grid, _("/// bad protocol : '%s'.\n"), cport.c_str());
    return;
  }

  int p = lookup_service(u.service);
  if (p == -1) {
    infof(conn->grid, _("/// bad port : '%s'.\n"), cport.c_str());
    return;
  }

  connect(conn, u.hostname.c_str(), p);
}

void cmd_match(conn_t *conn, int &quit) {
  my_wstring host = conn->buffer;
  if (host.find(' ')==-1)
    {
      set<my_wstring> nomatch;
      hl_matches = nomatch;
    }
  else {
    host = host.substr(host.find(' ')+1);
    if (hl_matches.find(host)==hl_matches.end())
      hl_matches.insert(host);
  }
  conn->grid->changed = 1;
}

void cmd_charset(conn_t *conn, int &quit) 
{
  my_wstring filename = conn->buffer;
  filename = filename.substr(filename.find(' ')+1);
  
  string cfilename = mks(filename);
  conn->mud_cset = cfilename;
  infof(conn->grid, _("/// charset '%s' selected\n"), conn->mud_cset.c_str());
}  

void cmd_dump(conn_t *conn, int &quit)
{
  my_wstring filename = conn->buffer;
  filename = filename.substr(filename.find(' ')+1);
  
  string cfilename = mks(filename);
  file_dump(conn->grid, cfilename.c_str());
}

void cmd_log(conn_t *conn, int &quit)
{
  my_wstring filename = conn->buffer;
  filename = filename.substr(filename.find(' ')+1);
  
  string cfilename = mks(filename);
  //  file_dump(cfilename.c_str());
  if (conn->logfile) {
    info(conn->grid, _("/// closing old logfile.\n"));
    fclose(conn->logfile);
    conn->logfile = NULL;
  }
  conn->logfile = fopen(cfilename.c_str(), "a");
  if (!conn->logfile) {
    infof(conn->grid, _("/// couldn't open '%s' for appending.\n"), cfilename.c_str());
  } else {
    infof(conn->grid, _("/// logging to end of '%s'.\n"), cfilename.c_str());
  }
}

struct {
  const wchar_t *commandname;
  void (*function)(conn_t *conn, int &quit);
} cmd_table[] =
{
#ifdef MCCP
  { L"compress", cmd_compress },
#endif
  { L"quit",     cmd_quit },
  { L"connect ", cmd_connect },
  { L"log ", 	 cmd_log },
  { L"match", 	 cmd_match },
  { L"close",    cmd_close },
  { L"open ",    cmd_connect },
  { L"dump ",    cmd_dump },
  { L"charset ", cmd_charset },
  { L"reload", cmd_reload },
  { L"help",     cmd_help },
  { L"z",        cmd_z },
  { NULL, }
};

void doenter(conn_t *conn, int &quit)
{
  
  if (commandmode) {
    chist()->insert(conn->buffer);

    commandmode = 0;
    int handled = 0;
    int i = 0;
    info(conn->grid, _("crystal> "));
    info(conn->grid, conn->buffer);
    info(conn->grid, "\n");

    const wchar_t *cmd = conn->buffer.data();
    int cmdlen = conn->buffer.length();

    wchar_t blah[cmdlen+2];
    wcsncpy(blah, conn->buffer.data(), cmdlen);
    blah[cmdlen] = 0;

    while (const wchar_t *cm=cmd_table[i].commandname) {
      if (wcsncmp(blah, cm, wcslen(cm))==0) {
	cmd_table[i].function(conn, quit);
	handled = 1;
	break;
      }
      i++;
    }
    if (!handled && conn->buffer.length()) {
      info(conn->grid, _("/// don't understand that command\n"));
    }
    conn->doclearline();
    return;
  }
  
  my_wstring wproper = L"";
  string proper = "";
  
  for (int i=0;i<conn->buffer.length();i++) {
    wchar_t sb = conn->buffer[i];

    wproper += sb;
    
    if (sb < 0x80) {
      proper += sb;
    }  else {

      static iconv_t i=0;
      static string cur_cset = "";
    
      if (!i || cur_cset != conn->mud_cset) {
	if (i)
	  iconv_close(i);
	i = iconv_open(conn->mud_cset.c_str(), "WCHAR_T");
	cur_cset = conn->mud_cset;
      }
      wchar_t ic = sb;
      char *ibuf = (char*)&ic;
      size_t ilen = 4;
      char o[11];
      char *obuf = o;
      size_t olen = 10;
      iconv(i, &ibuf, &ilen, &obuf, &olen);
      if (olen != 10) {
	*obuf = 0;
	proper += o;
      }
    }
  }

  if (conn->grid->cstoredprompt.length()) {
    for (int i=0;i<conn->grid->cstoredprompt.length();i++) {
      conn->grid->place(&conn->grid->cstoredprompt[i]);
    }
    conn->grid->cstoredprompt.erase();
  }

  conn->grid->lastprompt = 0;
    
  if (!(conn->telnet && conn->telnet->will_echo)) {
    if (wproper.length()) {
      chist()->insert(wproper);
    }

    my_wstring a;
    for (int i=0;i<conn->grid->col;i++) {
      a += conn->grid->get(conn->grid->row, i).ch;
    }
    /* 
    if (a == L"[Press return to continue or 'q' to abort]" && wproper==L"") {
      grid.eraseline(0);
      grid.col = 0;
      grid.row-=2;
    }
    */
    
    for (int i=0;i<wproper.length();i++) {
      cell_t c = mkchar(wproper[i]);
      conn->grid->place(&c);
    }
    conn->grid->wantnewline();
    conn->grid->changed = 1;
  }

  proper += "\r\n";
  if (conn->telnet) {
    conn->telnet->send(proper);
  }
  conn->doclearline();
}

void conn_t::doscrolldown()
{
  if (hardscroll) {
    hardscroll += 10;
  } 
  if (hardscroll > (grid->row - tty.HEIGHT))
    hardscroll = 0;	      
  grid->changed = 1;
}

void conn_t::doscrollup()
{
  if (grid->row < tty.HEIGHT)
    return;

  if (hardscroll) {
    hardscroll -= 10;
  } else {
    hardscroll = (grid->row - tty.HEIGHT) - 9;
  }
  if (hardscroll < 1)
    hardscroll = 1;
  grid->changed = 1;
}

void conn_t::docutfromhere() {
  cutbuffer=buffer.substr(c);
  buffer = buffer.substr(0, c);
}

void conn_t::docuttohere() {
  cutbuffer = buffer.substr(0, c);
  buffer = buffer.substr(c);
  c = 0;
}

void conn_t::dopaste() 
{
  buffer.insert(c, cutbuffer);
}

typedef void (conn_t::*keybinding_method_t)();

struct keybinding_t {
  const wchar_t *s;
  keybinding_method_t m;
  void (*f)(conn_t *conn, int &quit);
} keys[] = {
  { L"backspace", &conn_t::dobackspace },
  { L"tab",       &conn_t::findnext },

  { L"c-a",       &conn_t::dofirstchar },
  { L"c-b",       &conn_t::doprevchar },
  { L"c-c",       &conn_t::doclearline },
  { L"c-d",       &conn_t::dodelete },
  { L"c-e",       &conn_t::dolastchar },
  { L"c-f",       &conn_t::donextchar },
  { L"c-k",       &conn_t::docutfromhere },
  { L"c-l",       &conn_t::dorefresh },
  { L"c-n",       &conn_t::donexthistory },
  { L"c-p",       &conn_t::doprevhistory },
  { L"c-t",       &conn_t::dotranspose },
  { L"c-u",       &conn_t::docuttohere },
  { L"c-w",       &conn_t::dokillword },
  { L"c-y",       &conn_t::dopaste },
  { L"c-z",       NULL, cmd_z },

  { L"m-b",       &conn_t::doprevword },
  { L"m-f",       &conn_t::donextword },

  { L"return",    NULL, doenter },

  { L"up",        &conn_t::doprevhistory },
  { L"down",      &conn_t::donexthistory },
  { L"left",      &conn_t::doprevchar },
  { L"right",     &conn_t::donextchar },

  { L"home",      &conn_t::doscrollstart },
  { L"end",       &conn_t::doscrollend },

  { L"delete",    &conn_t::dodelete },

  { L"pagedown",  &conn_t::doscrolldown },
  { L"pageup",    &conn_t::doscrollup },
};

void conn_t::dispatch_key(const my_wstring &s, int &quit)
{
  if (s == L"c-]") {
    if (!commandmode) doclearline();
    commandmode = 1;
    return;
  }
  if (s.length()==1) {
    doinsertchar(s[0]);
    return;
  }
  
  if (s.length()>3 && s[0]=='f' && s[1]=='n' && s[2]=='.') {
    string cs = mks(s);
    triggerfn(cs.c_str()+3);
    return;
  }
  
  for (size_t i=0;i<(sizeof(keys)/sizeof(keybinding_t));i++) {
    if (keys[i].s == s) {
      if (keys[i].m)
	(this->*keys[i].m)();
      else
	keys[i].f(this, quit);
      return;
    }
  }
}

int had_winch = 0;

void winch(int) {
  tty.grabwinsize();
  had_winch = 1;
}

void connect(conn_t *conn, const char *host, int port)
{
  /* nuke old stuff */
  delete conn->telnet;
  conn->telnet = 0;

  if (conn->slave) conn->slave->visible = 0;
  if (conn->grid->col) 
    conn->grid->newline();

  infof(conn->grid, _("/// resolving %s\n"), host, port);
  display_buffer(conn);

  InAddrList *addrs = InAddr::resolv(host);
  if (!addrs) {
    infof(conn->grid, _("/// unknown host.\n"));
    return;
  }

  for (int i=0;i<addrs->size();i++) {
    InAddr *addr = addrs->get(i);

    infof(conn->grid, _("/// connecting to %s:%i\n"), addr->tostring().c_str(), port);
    display_buffer(conn);
    
    Socket *s2 = new Socket();
    addr->set_port(port);
    int stat = s2->connect(addr);
    //    delete addr;
    
    int e = errno;
    if (stat < 0) {
      infof(conn->grid, _("/// unable to connect : %s.\n"), strerror(e));
      delete s2;
      continue;
      return;
    }
    
    if (conn->telnet) delete conn->telnet;
    conn->telnet = new telnet_state(s2);
    
    conn->grid->cstoredprompt.erase();
    
    infof(conn->grid, _("/// connected\n"), host, port);
    
    static int a = 0;
    if (!a) {
      infof(conn->grid, _("/// escape character is '%s'\n"), "^]");
      a = 1;
    }
    
    tty.title(_("telnet://%s:%i - Crystal"), host, port);
    
    delete addrs;

    return;
  }
  
  delete addrs;
#if 0
  infof(conn->grid, _("/// exhausted all addresses\n"));
#endif
}

int file_dump(grid_t *grid, const char * file){
  FILE *dumpfile; 
  
  dumpfile = fopen(file,"a");
  if(NULL==dumpfile){
    infof(grid, _("/// couldn't open '%s' to dump to\n"), file);
    return 0;
  }
  infof(grid, _("/// dumping scroll history to '%s'\n"), file);
  
  time_t cur_time = time(NULL);
  string ctime_str = ctime(&cur_time);
  ctime_str = ctime_str.substr(0,ctime_str.length()-1);
    
  fprintf(dumpfile,_("---=== Dump generated on %s by crystal ===---\n"),
	   ctime_str.c_str());
  for(int i = grid->first;i<grid->row;i++){
    for(int j=0;j<grid->get_len(i);j++)
      fprintf(dumpfile,"%lc",grid->get(i,j).ch);
    fprintf(dumpfile,"\n");
  }
  fprintf(dumpfile,_("---=== End of dump ===---\n"));
  fclose(dumpfile);
  return 1;
}

void main_loop(conn_t *conn) 
{
  if (!conn->telnet)
    commandmode = 1;

  conn->grid->changed = 1;
  bad_have = 1;
  
  display_buffer(conn);
  
  int quit = 0;
  while (!quit) {
    fd_set r, e;

    int maxfd = 1;

    FD_ZERO(&r);
    FD_ZERO(&e);

    if (conn->telnet && conn->telnet->s) {
      FD_SET(conn->telnet->s->getfd(),&r);
      FD_SET(conn->telnet->s->getfd(),&e);

      maxfd = conn->telnet->s->getfd() + 1;
    }

    FD_SET(0, &r);

    struct timeval tm;
    struct timeval *tmp = &tm;
    if (conn->grid->nlw || scripting::count_timers()) {
      tm.tv_usec = 100000;
      tm.tv_sec = 0;
    } else {
      tmp = NULL;
    }

    select(maxfd, &r, 0, &e, tmp);

    scripting::dotimers();

    if (had_winch) {
      sendwinsize(conn);
      had_winch = 0;
      conn->grid->changed = 1;
      bad_have = 1;
    }
    
    for (int i=0;i<conn->grid->nlw;i++) {
      conn->grid->newline();
      conn->grid->changed = 1;
    }
    conn->grid->nlw = 0;

    if (conn->telnet && conn->telnet 
	&& (FD_ISSET(conn->telnet->s->getfd(), &r) || 
	    FD_ISSET(conn->telnet->s->getfd(), &e))) {
      int ok = 0;
      do {
	unsigned char mbuffer[1000];
	int bts = conn->telnet->s->read((char *)mbuffer, 500);
	ok = 0;
	if (bts == 0 || (bts==-1 && conn->telnet->s->getdead())) {
	  if (bts==-1) {
	    int e = errno;
	    infof(conn->grid, _("/// connection closed : %s.\n"), strerror(e));
	  } else {
	    info(conn->grid, _("/// connection closed by foreign host.\n"));
	  }
	  display_buffer(conn);
	  fflush(stdout);
	  ok = 0;
	  delete conn->telnet;
	  conn->telnet = 0;
	  commandmode = 1;
	  conn->grid->changed = 1;
	  break;
	} else if (bts!=-1) {
	  conn->telnet->decompress(conn, mbuffer, bts);
	  ok = 1;
	}
      } while (ok);
    }

    if (FD_ISSET(0, &r)) {
      
      while (1) {
	int foo;
	ioctl(0, FIONREAD, &foo);
	if(foo) {
	  wchar_t i = tty.getinput();

	  if (conn->telnet && conn->telnet->charmode && !commandmode
	      && i != 0x1d) {
	    string p;
	    p += i;
	    conn->telnet->send(p);
	  } else {
	    my_wstring s = tty.convert_input(i);
	    if (s.length())
	      conn->dispatch_key(s, quit);
	  }

	  if (!conn->telnet) {
	    commandmode = 1;
	  }
	  conn->grid->changed = 1;
	} else {
	  break;
	}
      }
    }
  display:
    if (!quit) {
      display_buffer(conn);
      fflush(stdout);
    }
  }
}

struct termios oldti;

int main(int argc, char **argv) {
  char *pname = argv[0];

  setlocale(LC_ALL, "");
  const char *codeset = nl_langinfo(CODESET);

  tty.getterm();

  grid_t grid(NULL);
  conn_t conn(&grid);
  grid.set_conn(&conn);
  
  scripting::set_grid(&grid);
  scripting::start();

  if (strcmp(codeset, "UTF-8")==0) {
    tty.utf8 = 1;
    conn.mud_cset = "UTF-8";
  }

#ifdef I18N
  bindtextdomain(PACKAGE, LOCALEDIR);
  textdomain(PACKAGE);
#endif
  if (argc>1) {
    url u = url(argv[1]);
    if (argc>2) {
      u.service = argv[2];
    }

    if (u.protocol != "telnet") {
      fprintf(stderr, _("%s: Bad protocol - '%s'.\n"), pname, u.protocol.c_str());
      exit(1);
    }

    int port = lookup_service(u.service);
    if (port == -1) {
      fprintf(stderr, _("%s: Bad port - '%s'.\n"), pname, u.service.c_str());
      exit(1);
    }

    connect(&conn, u.hostname.c_str(), port);
    if (!conn.telnet)
      exit(1);

    wchar_t blah[1000];
    if (argc>2)
      swprintf(blah, 1000, L"connect %s %s", argv[1], argv[2]);
    else
      swprintf(blah, 1000, L"connect %s", argv[1]);
    cmdhist.insert(blah);
  }

  info_to_stderr = 0;

  tty.grabwinsize();

  struct termios ti;
  
  tcgetattr(0, &oldti);
    
  cfmakeraw(&ti);
  tcsetattr(0, TCSADRAIN, &ti);

  setupterm(NULL, 0, NULL);
  signal(SIGWINCH, winch);

  main_loop(&conn);

  if (conn.logfile) {
    fclose(conn.logfile);
    conn.logfile = 0;
  }

  tcsetattr(0, TCSADRAIN, &oldti);
  printf("\n");
  return 0;
}
